#include <stdlib.h>
#include <string.h>
#include "nae-cfg.h"

#define NUM_GMAC_REGS        3

struct _per_cplx_ctl {
	uint32_t *gmac_regs;
	int *gmac_reg_id;
	int *loopback;
};

struct _per_cplx_ctl ctbl[NUM_NAE_COMPLX];

/* MAC_CONF1 */
#define INF_SOFTRESET(x)                    ((x) << 31)
#define INF_LOOP_BACK(x)                    ((x) << 8)
#define INF_RX_ENABLE(x)                    ((x) << 2)
#define INF_TX_ENABLE(x)                    (0x1)

/* MAC_CONF2 */
#define INF_PREMBL_LEN(x)                   (((x) & 0xf) << 12)
#define INF_IFMODE(x)                       (((x) & 0x3) << 8)
#define INF_LENCHK(x)                       ((((x) & 0x1)) << 4)
#define INF_PADCRCEN(x)                     (((x) & 0x1) << 2)
#define INF_PADCRC(x)                       (((x) & 0x1) << 1)
#define INF_FULLDUP(x)                      ((x) & 0x1)
#define TXINITIORCR(x)                      ((x) & 0x7ffff) << 8

static uint32_t format_reg(int intf, int reg)
{
	uint32_t fdt_val = ctbl[cur_complex].gmac_regs[intf + (reg * NUM_GMAC_REGS)];
	uint32_t val = 0;

	switch (reg) {

	case 1:
		val =	INF_PREMBL_LEN(0x7) |
			INF_IFMODE(0x2)     |
			INF_FULLDUP(1);
		return val;


	default:
		return fdt_val;
	};
}

static void init_gmac_regs(int num_regs)
{
	int nports = get_num_ports();

	ctbl[cur_complex].gmac_reg_id = (int *) malloc (num_regs * sizeof(int));
	assert(ctbl[cur_complex].gmac_reg_id != NULL);

	// Allocate total memory
	ctbl[cur_complex].gmac_regs = (uint32_t *) malloc(nports *  num_regs * sizeof(uint32_t));
	assert(ctbl[cur_complex].gmac_regs != NULL);

	ctbl[cur_complex].loopback = (int *) malloc(nports * sizeof(int));
	assert(ctbl[cur_complex].loopback != NULL);
}

void parse_gmac_regs(char *complex_str)
{
	int exp_len = get_num_ports();

	init_gmac_regs(NUM_GMAC_REGS);

	/* GMAC Regs */
	GET_FPROP(complex_str, "loopback_mode", ctbl[cur_complex].loopback);
	assert(prop_len == exp_len);
}

void gen_gmac_cfg(int intf)
{
	int j;
	int cur_id  = -1;
	uint32_t *gmac_regs = (uint32_t *) ctbl[cur_complex].gmac_regs;
	int *gmac_reg_id = (int *) ctbl[cur_complex].gmac_reg_id;
	int *loopback = (int *) ctbl[cur_complex].loopback;

	gmac_regs[intf] = INF_SOFTRESET(1) |
		INF_LOOP_BACK(loopback[intf]) |
		INF_RX_ENABLE(1) |
		INF_TX_ENABLE(1) ;
	gmac_reg_id[++cur_id] = 0;

	gmac_regs[intf + NUM_GMAC_REGS] = 0xdddddddd;
	gmac_reg_id[++cur_id] = 1;

	gmac_regs[intf + (2 * NUM_GMAC_REGS)] = INF_SOFTRESET(0) |
		INF_LOOP_BACK(loopback[intf]) |
		INF_RX_ENABLE(1) |
		INF_TX_ENABLE(1) ;
	gmac_reg_id[++cur_id] = 0;

	CFG_PRINT("\t\t\t\tintf-regs = <\n");
	for (j = 0; j < NUM_GMAC_REGS; ++j) {
		CFG_REG(gmac_reg_id[j], format_reg(intf, j));
	}
	CFG_PRINT("\t\t\t\t>;\n");

}

int  get_num_gmac_regs(void)
{
	return NUM_GMAC_REGS; /* for now */
}
