#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <libfdt.h>
#include <assert.h>

#include "nae-cfg.h"


/* Global Variables */
int total_num_ports;
int mode[NUM_NAE_COMPLX];
int num_ports[NUM_NAE_COMPLX];
int cur_mode;
int cur_complex;
int frequency;
int xaui_complex_map;

/*
 *  NAE Reg functions
 */
#if 0
static uint32_t iface_fifo_cfg(int iface, int sz)
{
	static int cur_start = 0;
	uint32_t reg = (iface & 0x1f) | (cur_start << 8) | (sz << 16);
	cur_start += sz;
	return reg;
}
#endif

/*
 *    FDT Parsing
 */
static void* fdt;
int fdt_err;
const void *pbuf;
char arcs_cmdline[2048];
int prop_len;
FILE* cfg_file;

/* Helper functions to parse FDT tree */
int get_fdt_prop(const char* node_path, const char *prop, void *prop_buf, int type)
{
	int i;
	char * src, *dst;

	/* Get the Node */
	int node = fdt_path_offset(fdt, node_path);
	if (node < 0) {
		printf("%s: Failed to parse node path %s \n",
			   fdt_strerror(node), node_path);
		return -1;
	}
	/* Get the Prop */
	pbuf = (char *) fdt_getprop(fdt, node, prop, &prop_len);
	if (pbuf == NULL) {
		printf("%s: Failed to parse %s \n", fdt_strerror(prop_len), prop);
		return -1;
	}
	if (type == PROP_STR) {
		memcpy((void *) prop_buf, (void * )pbuf, prop_len);

	} else { /* Copy in reverse to account for host LE order */
		src = (char *) pbuf;
		dst = (char *) prop_buf;
		for (i = 0; i < prop_len; ++i) {
			dst[i] = src[prop_len - i - 1];
		}
	}

	prop_len /= 4;
	printf("FDT: parsed %s.%s with len %d val: ", node_path, prop, prop_len);  
	if (type == PROP_INT) {
		for (i = 0; i < prop_len; ++i) {
			printf("%d ", *((unsigned int *) prop_buf + i));
		}
	} else {
		printf("%s ", (char *) prop_buf);
	}
	printf("\n");
	return 0;
}

static inline const char *basename(const char *path)
{
	const char *filename = strrchr(path, '/');
	return (filename) ? filename + 1 : path;
}


void parse_ucore()
{
#define FILE_BUFSZ 1024
        char file[FILE_BUFSZ];
	int ucore_mask;
	char cmd[1024];
	const char *uc_file = "ucore.dts";
	FILE *ucore_dts = fopen(uc_file, "w");
	FILE* hex_file;
	int i;
	char opcode[100];

#define UCFG_PRINT(args...) fprintf(ucore_dts, args)

	GET_FPROP_STR("/ucore/src@1", "file", file);
	/* Special handling for relative path */
	if (file[0] != '/') {
		char *filename = strdup(basename(file));
		if (!filename) {
			printf("ERROR: strdup failed\n");
			exit(1);
		}
		const char *ucore_dir = getenv("UCORE_DIR");
		int ret = snprintf(file, FILE_BUFSZ, "%s/%s", ucore_dir, filename);
		free(filename);
		if (ret > FILE_BUFSZ) {
			printf("ERROR: snprintf needs a buffer bigger than %d\n", FILE_BUFSZ);
			exit(1);
		}
	}
	if (access(file, R_OK) != 0) {
	    printf("ERROR: Ucore file %s Does not exist\n", file);
	    exit(1);
	}
	GET_FPROP("/ucore/src@1", "mask", &ucore_mask);

	sprintf(cmd, "make -C $UCORE_DIR UC00=%s -f Makefile.ucore_apps clean all",file);
	if (system(cmd) == -1) {
	    printf("ERROR: Failed to execute %s\n", cmd);
	    exit(1);
	}
	UCFG_PRINT("\t\t\tucore {\n");
	UCFG_PRINT("\t\t\t\tsrc@1 {\n");
	UCFG_PRINT("\t\t\t\t\tmask = <0x%x>;\n", ucore_mask);
	UCFG_PRINT("\t\t\t\t\topcodes = <\n");

	/* change the file suffix from .x to .hex */
	i = strlen(file);
 	file[i - 1] = 'h';
 	file[i]     = 'e';
 	file[i + 1] = 'x';
 	file[i + 2] = '\0';
	hex_file = fopen(file, "r");
	if (!hex_file) {
		fprintf(stderr, "Unable to open \"%s\" for reading! (%s)\n", file, strerror(errno));
		exit(-1);
	}
	while (fgets(opcode, 100, hex_file) != NULL) {
	    UCFG_PRINT("\t\t\t\t\t\t0x%s", opcode);
	}
	fclose(hex_file);

	UCFG_PRINT("\t\t\t\t\t>;\n");
	UCFG_PRINT("\t\t\t\t};\n");
	UCFG_PRINT("\t\t\t};\n");

	fclose(ucore_dts);
 	sprintf(cmd, "/bin/sed '/src@1/ a num-opcodes = <'\"`wc -l %s | cut -f1 -d \" \" `\"'>;' -i %s", file, uc_file);
	if (system(cmd) == -1) {
	    printf("ERROR: Failed to execute %s\n", cmd);
	    exit(1);
	}
}

static void parse_cfg(char *dev_cfg)
{
	struct stat st;
	int fd = open(dev_cfg, O_RDONLY);
	char prop[200];
	int i;
	char cplx_str[20];
	int il_num = 0;
	int offset;

	if (fd < 0) {
		perror("Failed to open cfg file ");
		exit(1);
	}
	if (fstat(fd, (struct stat *)&st) < 0) {
		perror(" Failed to read file status ");
		exit(1);
	}
	fdt = mmap(NULL, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	if (fdt ==  MAP_FAILED) {
		perror(" Failed to mmap cfg file ");
		exit(1);
	}
	fdt_err = fdt_check_header(fdt);
	if (fdt_err < 0) {
		printf("FDT_ERROR: Invalid FDT Detected: %s\n", fdt_strerror(fdt_err));
		exit(1);
	}
	printf("%s FDT @ %p is Validated\n", __func__, fdt);

	GET_FPROP("/", "frequency", &frequency);

	/* Common Config */
	parse_common_nae_regs();
	parse_poe_regs();

	/* Per Complex Config */
	for (i = 0; i < NUM_NAE_COMPLX; ++i) {
		sprintf(cplx_str, "/complex@%d", i);

		/* Check if the complex exists */
		offset = fdt_path_offset(fdt, cplx_str);
		if (offset == -FDT_ERR_NOTFOUND) {
			printf("Did not find %s\n", cplx_str);
			continue;
		}

		GET_FPROP_STR(cplx_str, "mode", prop);

		cur_complex = i;
		if (strncmp(prop, "sgmii", 5) == 0) {
			cur_mode = SGMII_MODE;
			mode[i] = SGMII_MODE;
	
			parse_gmac_regs(cplx_str);
			if (i == (NUM_NAE_COMPLX - 1)) {
				num_ports[i] = 2;
			} else {
				num_ports[i] = 4;
			}

		} else if (strncmp(prop, "xaui", 4) == 0) {
			cur_mode = XAUI_MODE;
			mode[i] = XAUI_MODE;
			parse_xaui_regs(cplx_str, 0);
			num_ports[i] = 1;
			xaui_complex_map |= ( 1 << i);

		} else if (strncmp(prop, "interlaken", 10) == 0) {
#if 0
			/* TODO: Read the num lanes and pass it on to calendering computation */
			/* This Needs to be fixed */
			if (i == 4) {
				printf("NAE_CFG_ERR: Interlaken should be configured in complex pair (0,1) or (2,3)\n");
				exit(0);
			}

			if ((i == 1) || (i == 3)) {
				if (il_num != 1) {
					printf("NAE_CFG_ERR: Illegal Interlaken Config detected\n");
					printf("Interlaken should be configured in complex pair (0,1) or (2,3)\n");
					exit(0);
				}
				continue; /* Skip complex 1 , 3 for IL */
			}
#endif
			cur_mode = INTERLAKEN_MODE;
			mode[i] = INTERLAKEN_MODE;
			parse_il_regs(cplx_str, 0);
			num_ports[i] = 1;
			il_num++;

		} else {
			printf("NAE_CFG_ERR: Unsupported mode %s detected for NAE complex%d\n", prop, i);
			exit(0);
		}
		total_num_ports += num_ports[i];
		parse_nae_regs(cplx_str);
			
	}

	CFG_PRINT("\t\t\tfrequency = <%d>;\n", frequency);
	CFG_PRINT("\t\t\txaui-complex-map = <0x%x>;\n", xaui_complex_map);
	parse_ucore();

	printf("%s for %s Done !\n", __func__, dev_cfg);
}

static char *get_mode_str(int mode)
{
	switch (mode) {
	case XAUI_MODE:
		return "XAUI_IF";

	case INTERLAKEN_MODE:
		return "INTERLAKEN_IF";
	default:
		return "SGMII_IF";
	};
}
void gen_cfg_dts()
{
	int i, j;
	int cur_port = 0;

	CFG_PRINT("\t\t\tnum-ports = <%d>;\n", total_num_ports);

	gen_global_nae_cfg();

	for (j = 0; j < NUM_NAE_COMPLX; ++j) {

		cur_complex = j;
		cur_mode = mode[j];

		for (i = 0; i < num_ports[j]; ++i) {
			CFG_PRINT("\t\t\tport@%d {\n", cur_port++);
			CFG_PRINT("\t\t\t\ttype = \"%s\";\n", get_mode_str(mode[j]));

			gen_nae_cfg(i);

			switch (mode[j]) {
			case SGMII_MODE:
				gen_gmac_cfg(i);
				break;

			case XAUI_MODE:
				gen_xaui_cfg(i);
				break;

			case INTERLAKEN_MODE:
				if ((j == 0) || (j == 2)) {
					gen_il_cfg(i);
				}
				break;

			default:
				break;
			};

			CFG_PRINT("\t\t\t};\n");
		}
	}

	gen_poe_config();

}

static void compute_cfg()
{
	compute_nae_config();
	compute_poe_config();
}

void init(const char* cfg_file_name)
{
	int start_port_id = 0;

	total_num_ports = 0;
	cfg_file = fopen(cfg_file_name, "w");
	if (cfg_file == NULL) {
		printf("Failed to Open NAE Config File %s\n", cfg_file_name);
		exit(-1);
	}
	CFG_PRINT("\t\tnae-cfg {\n");
	CFG_PRINT("\t\t\tstart-port-id = <%d>;\n", start_port_id);
	CFG_PRINT("\t\t\tnum-intf-regs = <%d>;\n", 0); /* Disabled temporarily */
}

#define CFG_FUNC_STR(func) STR(func)
#define STR(x) #x
void cleanup()
{
	CFG_PRINT("\t\t};\n");
	fclose(cfg_file);
}

#ifdef STANDALONE_NAE_CFG

int main(int argc, char**argv) {
	if (argc != 3) {
		printf("Invalid number of Arguments!\nUsage: %s <fdt.blob> <output.dts>\n", argv[0]);
		exit(-1);
	}
	init(argv[2]);
	parse_cfg(argv[1]);
	compute_cfg();
	gen_cfg_dts();
	cleanup();
	return 0;
}
#endif

int get_num_ports()
{
	switch (cur_mode) {
	case SGMII_MODE:
		return (cur_complex == 4) ? 2 : 4;

	case XAUI_MODE:
	case INTERLAKEN_MODE:
		return 1;

	default:
		return 4;
	};
}

int get_hw_port_id(int intf)
{
	switch (cur_mode) {
	case SGMII_MODE:
		if (cur_complex == 4) {
			return intf + 16;

		} else {
			return intf + (cur_complex * NUM_SGMII_PORTS);
		}

	case XAUI_MODE:
		return cur_complex * NUM_SGMII_PORTS;

	case INTERLAKEN_MODE:
		return (cur_complex / 2 )  * NUM_SGMII_PORTS * 2;

	default:
		return intf + (cur_complex * NUM_SGMII_PORTS);
	};
}
