#include <stdlib.h>
#include <string.h>
#include "nae-cfg.h"

#define MAX_FREE_DESC       1024
#define MAX_NAE_CONTEXTS    524
#define MAX_IACE_FIFO_SIZE (16 << 10) /* 16KB */
#define MAX_RX_BUF_SIZE     (256 << 10) /* 256 KB */
#define MAX_RX_BUF_ENTRIES  (MAX_RX_BUF_SIZE / 16) /* Each entry is 16Bytes */

#define MAX_CAL_SLOTS       64
#define SGMII_CAL_SLOTS     2
#define XAUI_CAL_SLOTS      13
#define IL8_CAL_SLOTS       20
#define IL4_CAL_SLOTS       10
#define FREQ_THRESHOLD      250
#define MAX_VFBID_ENTRIES   128
#define MAX_VC_DEST         (1 << 12)

#define MAX_FLOWS       (64 << 10) /* 64K flows */

#define NUM_NAE_REGS        2

extern int mode[NUM_NAE_COMPLX];

#define MAX_POE_CLASSES     8
#define MAX_POE_CLASS_CTXT_TBL_SZ ((MAX_NAE_CONTEXTS / MAX_POE_CLASSES) + 1)

enum {
	RX_IF_BASE_CONFIG0         = 0x12,
	RX_CAL_SLOT_CFG            = 0x1d,
	PARSER_SEQ_FIFO_CFG        = 0x1f,
	FREE_IN_FIFO_CFG           = 0x20,
	RX_BUF_BASE_DEPTH_ADDR     = 0x21,
	RX_BUF_BASE_DEPTH          = 0x22,
	FREE_IN_FIFO_UNIQ_CFG_SIZE = 0x28,
	FLOW_CRC16_POLY_CFG        = 0x2f,
	FLOW_BASE_MASK_CFG         = 0x80,
	VFBID_DESTMAP_CMD          = 0x3a,
	POE_CLASS_SETUP_CFG        = 0x81,
	UCORE_IFACE_MASK_CFG       = 0x82,
	IFACE_FIFO_CFG             = 0x8a,
};

struct _per_cplx_nae_ctl {
	uint32_t *nae_regs;
	int *nae_reg_id;
	int valid;
	int num_free_desc[4];
	int rx_buf_sizes[4];
	int intf_fifo_sizes[4];
	int num_channels[4];
	int prsr_seq_fifo_sz[4];
};

struct _common_nae_ctl {
	int cal_len;
	int num_regs;
	int total_free_desc;
};

struct _vfbid_tbl {
	int tbl[MAX_VFBID_ENTRIES];
	int n_entries;
};

static	int cur_iface_start;
static  int free_desc_reg_id;

static  int flow_crc_poly;
static  int cur_flow_base;
static  int per_port_num_flows;

static int cur_iface_start;

static struct _per_cplx_nae_ctl ctbl[NUM_NAE_COMPLX];
static struct _common_nae_ctl ctl;
static 	int cur_slot = 0;
static  struct _vfbid_tbl vfbid;

static uint32_t get_fdt_reg(int intf, int reg)
{
	int num_ports = get_num_ports();
	return ctbl[cur_complex].nae_regs[intf + (reg * num_ports)];
}
static uint32_t format_reg(int intf, int reg)
{
	uint32_t fdt_val = get_fdt_reg(intf, reg);
	int hw_port = get_hw_port_id(intf);

	switch (ctbl[cur_complex].nae_reg_id[reg]) {

	case FREE_IN_FIFO_UNIQ_CFG_SIZE: 
	{
		int size_in_clines =  fdt_val / CACHE_LINE_SIZE;
		return ((size_in_clines << 8) | (hw_port & 0x1f)); 
	}

	case UCORE_IFACE_MASK_CFG: 
	{
		return ( 0x1ULL << 31) | (((fdt_val & 0xffff)  << 8) | (hw_port & 0x1f)); 
	}

	default:
		return fdt_val;
	};
}

static void init_nae_regs(int num_regs)
{
	int nports = get_num_ports();

	ctbl[cur_complex].nae_reg_id = (int *) malloc (num_regs * sizeof(int));
	assert(ctbl[cur_complex].nae_reg_id != NULL);

	// Allocate total memory
	ctbl[cur_complex].nae_regs = (uint32_t *) malloc(nports *  num_regs * sizeof(uint32_t));
	assert(ctbl[cur_complex].nae_regs != NULL);

	ctbl[cur_complex].valid = 1;
}

void parse_nae_regs(char *complex_str)
{
	unsigned char *rbuf = NULL;
	int reg_blk_sz = 0;
	int cur_nae_reg_id = -1;
	int num_ports = get_num_ports();
	int *nae_reg_id = NULL;
	int i, tot_sz = 0;
	int *num_channels = (int *) ctbl[cur_complex].num_channels;

	/* Interlaken is not supported yet */
	if (cur_mode == INTERLAKEN_MODE) {
		return;
	}

	init_nae_regs(NUM_NAE_REGS);

	rbuf = (unsigned char *) ctbl[cur_complex].nae_regs;
	reg_blk_sz = num_ports * sizeof (uint32_t);
	nae_reg_id = ctbl[cur_complex].nae_reg_id;


	GET_FPROP(complex_str, "iface_fifo_sizes", ctbl[cur_complex].intf_fifo_sizes);
	assert(prop_len == num_ports);
 
	GET_FPROP(complex_str, "parser_sequence_fifo_sizes", ctbl[cur_complex].prsr_seq_fifo_sz);
	assert(prop_len == num_ports);

	GET_FPROP(complex_str, "num_channels", num_channels);
	assert(prop_len == num_ports);
   
	for (i=0; i < num_ports; ++i) {
		tot_sz += num_channels[i];
	}

	GET_FPROP(complex_str, "rx_buffer_sizes", ctbl[cur_complex].rx_buf_sizes);
	assert(prop_len == tot_sz);

	GET_FPROP(complex_str, "free_desc_sizes", rbuf);
	assert(prop_len == num_ports);
	nae_reg_id[++cur_nae_reg_id] = FREE_IN_FIFO_UNIQ_CFG_SIZE;
	free_desc_reg_id = cur_nae_reg_id;
	rbuf += reg_blk_sz;

	GET_FPROP(complex_str, "ucore_mask", rbuf);
	assert(prop_len == num_ports);
	nae_reg_id[++cur_nae_reg_id] = UCORE_IFACE_MASK_CFG;

	GET_FPROP(complex_str, "num_free_descs", ctbl[cur_complex].num_free_desc);
	assert(prop_len == num_ports);

}

#define MAX_SGMII_PORTS 18
static void gen_freein_fifo_cfg()
{
	int size, i, cp = 0;
	uint32_t reg;
	int start = 0;


	for (cp = 0 ; cp < NUM_NAE_COMPLX; ++cp ) {
		for (i = 0; i < 4; ++i) { /* 4 interfaces */

			/* Each entry represents 2 descs; hence division by 2 */
			size = ctbl[cp].num_free_desc[i] / 2;
			if (size == 0 ) size = 8;

			reg = ((size  & 0x3ff ) << 20) | /* fcSize */
				((start & 0x1ff)  << 8) | /* fcStart */
				(((cp * 4) + i)  & 0x1f);

			CFG_REG(FREE_IN_FIFO_CFG, reg);
			start += size;
		}
	}
}
/* Compute data to be used while generating CFG files.
 */
void compute_nae_config()
{
	int i, j;
	int n_sgmii_slots = SGMII_CAL_SLOTS * ((frequency <= FREQ_THRESHOLD) ? 2 : 1);
	int n_xaui_slots = XAUI_CAL_SLOTS * ((frequency <= FREQ_THRESHOLD) ? 2 : 1);
	int n_ports;

	ctl.num_regs += 10; /* For Rx Base Config */
	ctl.num_regs += 60; /* For Rx Buf Config */
	ctl.num_regs += 20; /* For Free Fifo Config */
	ctl.num_regs += 20; /* For Interface Fifo Config */
	ctl.num_regs += 20; /* For Praser Seq Fifo Config */

	for (i = 0; i < NUM_NAE_COMPLX; ++i) {
		cur_mode = mode[i];
		cur_complex = i;
	
		if (ctbl[i].valid == 0) {
			continue;
		}

		/* RX calendering */
		switch (mode[i]) {
		case SGMII_MODE:
			/* Even the complex 4 gets 4 slots */
			ctl.cal_len += 4 * n_sgmii_slots; 
			break;

		case XAUI_MODE:
			ctl.cal_len += n_xaui_slots;
			break;

		case INTERLAKEN_MODE:
			if (i % 2) {
				continue;
			}
			ctl.cal_len += IL4_CAL_SLOTS; /* For now. To be parsed from FDT */
			break;

		default:
			break;

		};

		/* Free Fifo Carving */

		n_ports = get_num_ports();
		for (j = 0; j < n_ports; ++j) {
			ctl.total_free_desc += ctbl[i].num_free_desc[j];
		}
	}

	assert(ctl.total_free_desc <= MAX_FREE_DESC);
	assert(ctl.cal_len <= MAX_CAL_SLOTS);
	per_port_num_flows = MAX_FLOWS / total_num_ports;
	
}
/*
 */
static void gen_cal_cfg(intf)
{
	int k;
	uint32_t reg;
	int n_sgmii_slots = SGMII_CAL_SLOTS * ((frequency <= FREQ_THRESHOLD) ? 2 : 1);
	int n_xaui_slots = XAUI_CAL_SLOTS * ((frequency <= FREQ_THRESHOLD) ? 2 : 1);
	static int cur_slot_base = 0;
	switch (cur_mode) {
	case SGMII_MODE:
		for (k = 0; k < n_sgmii_slots; ++k) {

			reg = (ctl.cal_len << 16) | (cur_slot++); 
			reg |= (((cur_complex * 4) + intf) << 8);
			CFG_REG(RX_CAL_SLOT_CFG, reg);
			/* printf("\t\t\t\t\t0x%x 0x%08x\n", */
			/* 	  RX_CAL_SLOT_CFG, reg); */
		}
		break;

	case XAUI_MODE:
		for (k = 0; k < n_xaui_slots; ++k) {

			reg = (ctl.cal_len << 16) | (cur_slot_base + (k * 4));
			reg |= ((cur_complex * 4) << 8);
			CFG_REG(RX_CAL_SLOT_CFG, reg);
			/* printf("\t\t\t\t\t0x%x 0x%08x\n", */
			/*        RX_CAL_SLOT_CFG, reg); */
		}
		cur_slot = cur_slot_base + ((k - 1) * 4) + 1;
		break;

	case INTERLAKEN_MODE:
		if (cur_complex % 2) {
			break;
		}
		/* TODO: Fix This */
		for (k = 0; k < IL4_CAL_SLOTS; ++k) {

			reg = (ctl.cal_len << 16) | (++cur_slot);
			reg |= ((cur_complex * 4) << 8);
			CFG_REG(RX_CAL_SLOT_CFG, reg);
			/* printf("\t\t\t\t\t0x%x 0x%08x\n", */
			/*        RX_CAL_SLOT_CFG, reg); */
		}
		break;

	default:
		break;

	};
	cur_slot_base++;
}
static int get_flow_mask(int num_ports)
{
	int i;
	int max_bits = 5; /* upto 32 ports */
	/* Compute the number of bits to needed to
	 * represent all the ports */
	for (i = 0; i < max_bits; ++i) {
		if (num_ports <= (2 << i)) {
			return i + 1; /* num bits is non zero and 'i' starts from 0 */
		}
	}
	return max_bits;
}
/* Flow Config */
static void gen_flow_cfg(int intf)
{
	int port = get_hw_port_id(intf);
	int flow_mask = get_flow_mask(total_num_ports);
	uint32_t reg = (cur_flow_base << 16) | port;
	reg |= ((flow_mask & 0x1f) << 8);

	CFG_REG(FLOW_BASE_MASK_CFG, reg);
	cur_flow_base += per_port_num_flows;
}

int get_num_nae_regs(int intf)
{
	int cal_regs = SGMII_CAL_SLOTS * ((frequency <= FREQ_THRESHOLD) ? 2: 1);
	return NUM_NAE_REGS + 1 + cal_regs ; /* Hack ! For now. To be fixed */
}


static int get_num_nae_regs_per_port(int intf)
{
	/* Curretnly Interlaken is not supported */
	if (cur_mode == XAUI_MODE) {
		int cal_regs = XAUI_CAL_SLOTS * ((frequency <= FREQ_THRESHOLD) ? 2 : 1);
		return 	NUM_NAE_REGS + 1 +
			cal_regs; /* Hack ! For now. To be fixed */

	} else {
		return get_num_nae_regs(intf);
	}
}
/* RX Buf Carving
 */
static void gen_rx_buf_cfg()
{
	uint32_t val;
	int i, sz, cp, j;
	int *num_channels = (int *) ctbl[cur_complex].num_channels;
	int context = 0;
	int base = 0;
	int nc = 0;


	for (cp = 0; cp < NUM_NAE_COMPLX; ++cp){
		num_channels = (int *) ctbl[cp].num_channels;

		for (j = 0; j < 4; ++j) {
			nc = (num_channels[j]) ? (num_channels[j]) : 1;

			for (i = 0; i < nc; ++i) {

				/* Write Index (context num) */
				CFG_REG(RX_BUF_BASE_DEPTH_ADDR, context + i);

				/* Write val (rx buf sizes) */
				sz = ctbl[cp].rx_buf_sizes[j];
				val = 0x80000000 | ((base << 2) & 0x3fff); /* base */
				val |= (((sz << 2)  & 0x3fff) << 16); /* size */

				CFG_REG(RX_BUF_BASE_DEPTH, val);
				CFG_REG(RX_BUF_BASE_DEPTH, 0x7fffffff & val);
				base += sz;

			}
			context += nc;
		}
	}
}

void gen_nae_cfg(int intf)
{
	int j;
	int tx_qid = get_hw_port_id(intf) + NAE_TX_VC_BASE;
	int rx_qid = get_hw_port_id(intf) + NAE_RX_VC_BASE;

	CFG_PRINT("\t\t\t\tnum-free-desc = <%d>;\n", ctbl[cur_complex].num_free_desc[intf]);
	CFG_PRINT("\t\t\t\thw-port-id = <%d>;\n", get_hw_port_id(intf));
	CFG_PRINT("\t\t\t\tmgmt = <%d>;\n", (cur_complex == 4) ? 1 : 0);
	CFG_PRINT("\t\t\t\ttx-que-range = <%d %d>;\n", tx_qid, tx_qid);
	CFG_PRINT("\t\t\t\trx-que = <%d>;\n", rx_qid);
	CFG_PRINT("\t\t\t\tnum-nae-regs = <%d>;\n", get_num_nae_regs_per_port(intf));

	if (ctbl[cur_complex].valid == 0) {
		return;
	}

	CFG_PRINT("\t\t\t\tfree-desc-size = <%d>;\n", get_fdt_reg(intf, free_desc_reg_id));

	CFG_PRINT("\t\t\t\tnae-regs = <\n");

	for (j = 0; j < NUM_NAE_REGS; ++j) {
		CFG_REG(ctbl[cur_complex].nae_reg_id[j],
			format_reg(intf, j));
	}


	/* 2. Generate cfg for all the non-parsed regs */

	/* RX calendering */
	gen_cal_cfg(intf);

	/* FLOW base config */
	gen_flow_cfg(intf);

	CFG_PRINT("\t\t\t\t>;\n");
}

void parse_common_nae_regs(void)
{
	GET_FPROP("/", "vfbid-map", vfbid.tbl);
	assert(prop_len <= (2 * MAX_VFBID_ENTRIES));

	/* Each entry is a pair of (vfbid, dest map) */
	vfbid.n_entries = prop_len / 2;
	ctl.num_regs +=  vfbid.n_entries;

	/* Flow CRC16 Poly */
	GET_FPROP("/", "flow_crc_poly", &flow_crc_poly);
	assert(prop_len  == 1);
	ctl.num_regs++;
}

static void gen_poe_class_cfg()
{
	int i;
	uint32_t val;
	int poe_cl_tbl[MAX_POE_CLASSES] = {0x0, 0x249249, 0x492492,
					   0x6db6db, 0x924924, 0xb6db6d,
					   0xdb6db6, 0xffffff};

	for (i = 0; i <MAX_POE_CLASS_CTXT_TBL_SZ; ++i) {
		val = (0x80 | i); /* set the write bit 0x80 */
		val |= (poe_cl_tbl[(i/MAX_POE_CLASSES) & 0x7] << 8);

		CFG_REG(POE_CLASS_SETUP_CFG, val);
	}
}

static void gen_vfbid_mapping()
{
	int i;
	uint32_t reg;
	for (i=0; i < vfbid.n_entries; ++i) {
		assert(vfbid.tbl[2*i + 1] <= (MAX_VFBID_ENTRIES - 1)); 
		reg = (vfbid.tbl[2*i + 1] << 4) | 1;
		reg |= (vfbid.tbl[2*i] << 16);
		assert(vfbid.tbl[2*i] < (MAX_VC_DEST)); 
		CFG_REG(VFBID_DESTMAP_CMD, reg);
	}

}

static void gen_iface_fifo_cfg()
{
	int fifo_xoff_thresh = 12;
	uint32_t reg;
	int i, size, j;
	for (i = 0; i < NUM_NAE_COMPLX; ++i) {

		for (j = 0 ; j <4; ++j) {
			size = ctbl[i].intf_fifo_sizes[j];
			reg = ((fifo_xoff_thresh << 25) |
			       ((size & 0x1ff) << 16) |
			       ((cur_iface_start & 0xff) << 8) |
			       (((i * 4) + j) & 0x1f)); 
			CFG_REG(IFACE_FIFO_CFG, reg);
			cur_iface_start += size;
		}
	}
}

static void gen_rx_base_cfg()
{
	int i, j;
	int id;
	int *num_channels;
	int base = 0;
	uint32_t val, nc;

	for (i = 0; i < NUM_NAE_COMPLX; ++i) {

		num_channels = (int *) ctbl[i].num_channels;
		id = RX_IF_BASE_CONFIG0 + (i * 2);
	
		for (j=0; j < 2; ++j) {
			val = (base & 0x3ff);
			nc = (num_channels[2*j]) ? num_channels[2*j] : 1;
			base += nc;

			val |= ((base & 0x3ff) << 16);
			nc = (num_channels[2*j + 1]) ? num_channels[2*j + 1] : 1;
			base += nc;

			CFG_REG(id + j, val);
		}
	}
}

static void gen_parser_seq_fifo_cfg()
{
	uint32_t val;
	int start = 0, size, i, j;
	int *sizes;

	for (i = 0; i < NUM_NAE_COMPLX; ++i) {

		sizes = (int *) ctbl[i].prsr_seq_fifo_sz;
		
		for (j=0; j < 4; ++j) { 
			size = sizes[j];

			val = (((size & 0x1fff) << 17) |
			       ((start & 0xfff) << 5) |
			       (((i * 4) + j) & 0x1f)); 
			CFG_REG(PARSER_SEQ_FIFO_CFG, val);
			start += size;
		}
	}
}

void gen_global_nae_cfg()
{
	uint32_t reg;
	
	CFG_PRINT("\t\t\tglobal-nae-regs {\n");
	CFG_PRINT("\t\t\t\t num-regs = <0x%x>;\n",
		  MAX_POE_CLASS_CTXT_TBL_SZ + ctl.num_regs);
	CFG_PRINT("\t\t\t\t regs = <\n");

	/* 1. POE_CLASS_SETUP_CFG */
	gen_poe_class_cfg();

	/* 2.  VFBID  */
	gen_vfbid_mapping();

	/* 3. FLOW CRC16 Poly */
	reg = flow_crc_poly;
	CFG_REG(FLOW_CRC16_POLY_CFG, reg);

	/* 4. IFACE_FIFO_CFG  */
	gen_iface_fifo_cfg();

	/* 5. RX Base Config */
	gen_rx_base_cfg();

	/* 6. RX Buffers */
	gen_rx_buf_cfg();

	/* 7. Freein FIFO Config */
	gen_freein_fifo_cfg();

	/* 8. Parser Seq Fifo */
	gen_parser_seq_fifo_cfg();

	CFG_PRINT("\t\t\t\t >;\n");
	CFG_PRINT("\t\t\t};\n");
}
