#include <stdlib.h>
#include <string.h>
#include "nae-cfg.h"

enum {
	POE_PCI_CFG_REG = 0,
	POE_PCI_MEM_REG = 1,
};
#define CFG_PCIMEM_REG(index, val) CFG_PRINT("\t\t\t\t\t%d 0x%x 0x%08x\n", POE_PCI_MEM_REG, (index), (val))
#define CFG_PCICFG_REG(index, val) CFG_PRINT("\t\t\t\t\t%d 0x%x 0x%08x\n", POE_PCI_CFG_REG, (index), (val))

#define POE_DV_NODE "/poe/distribution_vectors"

#define NUM_DIST_VEC 16
#define NUM_WORDS_PER_DV 16
#define MAX_DV_TBL_ENTRIES (NUM_DIST_VEC * NUM_WORDS_PER_DV)
#define POE_DIST_THRESHOLD_0 0x200
#define POE_DIST_ENABLE 0x205
#define POE_DIST_VEC0_512 0x100
#define POE_DIST_THRESHOLD_VAL 0xa

static unsigned int dvtbl[MAX_DV_TBL_ENTRIES];
static int num_regs;

void parse_poe_regs()
{
	int i;
	char str[50];
	unsigned char *buf = (unsigned char *) dvtbl;

	for (i=0; i < NUM_DIST_VEC; ++i) {
		
		sprintf(str, "dv%d", i);
		GET_FPROP(POE_DV_NODE, str, buf);
		assert(prop_len == NUM_WORDS_PER_DV);

		buf += (NUM_WORDS_PER_DV * 4);
	}
#if 0
	for (i=0; i < NUM_DIST_VEC; ++i) {
		for (j=0; j < NUM_WORDS_PER_DV; ++j) {
			printf("%d ", dvtbl[(i * NUM_WORDS_PER_DV)  + NUM_WORDS_PER_DV - j - 1]);
		}
		printf("\n");
	}
#endif
}

void gen_poe_config()
{
	int i, j;
	int reg_id = POE_DIST_VEC0_512;
	uint32_t val;

	CFG_PRINT("\t\t\tpoe-pci-mem-regs {\n");
	CFG_PRINT("\t\t\t\tnum-regs = <%d>;\n", num_regs);
	CFG_PRINT("\t\t\t\tregs = <\n");
	/* Disable  DIST VEC logic */
	CFG_PCICFG_REG(POE_DIST_ENABLE, 0); 

	for (i=0; i < NUM_DIST_VEC; ++i) {
		for (j=0; j < NUM_WORDS_PER_DV; ++j) {

			val = dvtbl[(i * NUM_WORDS_PER_DV)  + NUM_WORDS_PER_DV - j - 1];
			CFG_PCIMEM_REG(reg_id, val);
			reg_id += 1;
		}
	}
	/* Set the threshold */
	for (i=0; i< 5; ++i) {
		CFG_PCICFG_REG((i + POE_DIST_THRESHOLD_0), POE_DIST_THRESHOLD_VAL);
	}

	/* Enable DIST VEC logic */
	CFG_PCICFG_REG(POE_DIST_ENABLE, 1); 

	CFG_PRINT("\t\t\t\t>;\n");
	CFG_PRINT("\t\t\t};\n");
}

void compute_poe_config()
{
	/* Currently only Distribution vectors are parsed */
	num_regs = MAX_DV_TBL_ENTRIES;

	/* To Account for extra reg writes
	 * 1. disable DIST_VEC_ENABLE bit
	 * write all the dist vectors
	 * 2. Write threshold vectors (5 regs)
	 * 3. enable DIST_VEC_ENABLE bit
	 */ 
	num_regs += 7; 
}
